/*******************************************************************************
*    Robot "Giskard" 1.0.                                                      *
*    Copyright (C) 2009 Luca Bertoncello                                       *
*    Hartigstrasse, 12 - 01127 Dresden Deutschland                             *
*    E-Mail: lucabert@lucabert.de, lucabert@lucabert.com                       *
*    http://www.lucabert.de/  http://www.lucabert.com/                         *
*                                                                              *
*    Circuits  by Marian Neubert     marian@tesla-crew.de                      *
*    Mechanism by Mario Zimmermann   mario@tesla-crew.de                       *
*                                                                              *
*    This program is free software; you can redistribute it and/or modify      *
*    it under the terms of the GNU General Public License as published by      *
*    the Free Software Foundation; version 2 of the License.                   *
*                                                                              *
*    This program is distributed in the hope that it will be useful,           *
*    but WITHOUT ANY WARRANTY; without even the implied warranty of            *
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
*    GNU General Public License for more details.                              *
*                                                                              *
*    You should have received a copy of the GNU General Public License         *
*    along with this program; if not, write to the Free Software               *
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA *
*******************************************************************************/

#include <stdlib.h>
#include <string.h>
#include <avr/interrupt.h>
#include "inc/giskard.h"
#include "inc/tools.h"
#include "inc/leds.h"
#include "inc/voice.h"
#include "inc/twislave.h"

uint8_t isfirstbyte, hasabit, answerType;
boolean vdCommandType, canSendCommand, waitForAnswer;

extern volatile uint8_t rxbuffer[buffer_size], txbuffer[buffer_size], buffer_adr;

/**
* Funktion um eine komplette Initialisierung, inklusiv alle Prüfungen, des
* Sprachsystems
*/
void doInitVoice()
{
  uint8_t data[10];

  VD_DDR_MHS |= (1<<VD_MHS);
  VD_DDR_DATA |= (1<<VD_DATA);

  VD_DDR_SHS &= ~(1<<VD_SHS);
  VD_DDR_BSY &= ~(1<<VD_BSY);

  VD_PORT_MHS |= (1<<VD_MHS);
  VD_PORT_DATA |= (1<<VD_DATA);
  VD_PORT_BSY |= (1<<VD_BSY);

  VD_DDR_RST |= (1<<VD_RST);
  VD_PORT_RST &= ~(1<<VD_RST);
  delayMS(1);
  VD_PORT_RST |= (1<<VD_RST);

  vdIsBusyWait();

  isfirstbyte = 0;
  hasabit = 0;

  bzero(curMask, VD_MAX_WORDS);
  countCurMask = 0;

  // Mask setzen: alles erlauben
  data[0] = 0x04;
  data[1] = 0x29;
  data[2] = 0xFF;
  data[4] = 0x01;
  vdSendData(data);
  delayMS(20);
  vdReceiveData(data);

  vdCommandToSend = VD_CMD_NOOP;

  hasToSendCommand = TRUE;
  canSendCommand = TRUE;
  waitForAnswer = FALSE;
}

/**
* Berechnet die Prüfsumme der angegebenen Daten
*
* @param uint8_t   *data    Die Daten
* @return uint16_t          Die Prüfsumme
*/
uint8_t vdChecksum(uint8_t *data)
{
  uint8_t count;
  uint16_t check = 0;

  for(count = 1; count <= data[1]; count++)
    check += data[count];

  check %= 256;

  return(check);
}

/**
* Schickt einen Byte dem VoiceDirect
*
* @param uint8_t   data    Der zu sendenden Byte
*/
void vdSendByte(uint8_t data)
{
  int8_t count = 7;

  turnOnLED(LED4);

  VD_DDR_DATA |= (1<<VD_DATA);
  VD_PORT_MHS |= (1<<VD_MHS);
  for(count = 7; count >= 0; count--)
  {
    waitForPINSetWithTimeOut(VD_PIN_SHS, VD_SHS, 2000); // Eine Sekunde TimeOut
    if((data>>count) & 1)
      VD_PORT_DATA |= (1<<VD_DATA);
    else
      VD_PORT_DATA &= ~(1<<VD_DATA);
    VD_PORT_MHS &= ~(1<<VD_MHS);
    waitForPINClearWithTimeOut(VD_PIN_SHS, VD_SHS, 2000); // Eine Sekunde TimeOut
    VD_PORT_MHS |= (1<<VD_MHS);
  }
  VD_PORT_DATA |= (1<<VD_DATA);

  turnOffLED(LED4);
}

/**
* Empfängt einen Byte vom VoiceDirect
*
* @return uint8_t   Der empfangenen Byte
*/
uint8_t vdReceiveByte()
{
  int8_t count = 7, start, c = 7;
  uint8_t data = 0;

  turnOnLED(LED2);

  if(isfirstbyte == 0)
    start = 127;
  else
  {
    if(hasabit == 0)
    {
      start = 7;
      c = 7;
    }
    else
    {
      start = 6;
      c = 6;
      hasabit = 0;
    }
  }
  VD_DDR_DATA &= ~(1<<VD_DATA);

  for(count = start; count >= 0; count--, c--)
  {
    delayMS(2);
    waitForPINClearWithTimeOut(VD_PIN_SHS, VD_SHS, 2000); // 2 Sekunden TimeOut
    if(VD_PIN_DATA & (1<<VD_DATA))
      data |= (1<<c);
    VD_PORT_MHS &= ~(1<<VD_MHS);
    delayMS(2);
    waitForPINSetWithTimeOut(VD_PIN_SHS, VD_SHS, 2000); // 2 Sekunden TimeOut
    VD_PORT_MHS |= (1<<VD_MHS);
    if(isfirstbyte == 0 && !(VD_PIN_DATA & (1<<VD_DATA)))
    {
      count = -1;
      isfirstbyte = 1;
      hasabit = 1;
    }
  }

  turnOffLED(LED2);

  return data;
}

/**
* Empfängt einen Datenpaket vom VoiceDirect
*
* @param uint8_t   *data    Die empfangenen Bytes
* @return                   TRUE wenn alles in Ordnung,
*                           FALSE wenn die Prüfsumme nicht stimmt
*/
boolean vdReceiveData(uint8_t *data)
{
  int8_t count;
  uint8_t checksum = 0;

  isfirstbyte = 0; // beim ersten Mal MÜSSEN wir die Variabel auf 0 setzen!!

  data[0] = vdReceiveByte();
  data[1] = vdReceiveByte();

  for(count = 2; count <= (data[1] + 1); count++)
    data[count] = vdReceiveByte();
  checksum = vdChecksum(data);

  if(checksum != data[count - 1])
    return FALSE;
  return TRUE;
}

/**
* Schickt einen Paket dem VoiceDirect
*
* @param uint8_t   *data    Die Bytes zu senden. Der erste Byte (0) enthält die Länge des Pakets
*/
void vdSendData(uint8_t *data)
{
  uint8_t l;
  uint16_t check = 0;

  vdSendByte(0xFF);
  for(l = 0; l < data[0]; l++)
  {
    vdSendByte(data[l]);
    check += data[l];
  }
  check %= 256;
  vdSendByte(check);
}

/**
* Schickt einem Befehl dem VoiceDirect
*
* @param uint8_t  cmd     Der zu sendenden Befehl
*/
void sendCommand(uint8_t cmd)
{
  uint8_t data[10];
  uint8_t count, p, l;
  char sTemp[20];

  bzero(data, 10);

  switch(cmd)
  {
    case VD_CMD_LEARN:
      data[0] = 0x05;
      data[1] = 0x02;
//      data[2] = 0x79;  // Es wird gesagt: "Sagen Sie ein Wort"
//      data[3] = 0x00;
      data[2] = 0xDC; // low beep
      data[3] = 0x01;
      data[4] = 0x01;
      vdCommandType = CMDTYPE_RUN;
      waitForAnswer = TRUE;
      answerType = ANSWERTYPE_LEARN;
    break;
    case VD_CMD_HEAR:
      data[0] = 0x03;
      data[1] = 0x4C;
      data[2] = 0xFF;
      vdCommandType = CMDTYPE_RUN;
      waitForAnswer = TRUE;
      answerType = ANSWERTYPE_HEAR;
    break;
    case VD_CMD_HEAR_PROMPT:
      data[0] = 0x05;
      data[1] = 0x10;
      data[2] = 0x79;
      data[3] = 0x00;
      data[4] = 0x01;
      vdCommandType = CMDTYPE_RUN;
      waitForAnswer = TRUE;
      answerType = ANSWERTYPE_HEAR;
    break;
    case VD_CMD_DELETEDICT:
      data[0] = 0x03;
      data[1] = 0x28;
      data[2] = 0x55;
      vdCommandType = CMDTYPE_SYSTEM;
      waitForAnswer = TRUE;
      answerType = ANSWERTYPE_SYSTEM;
    break;
    case VD_CMD_GETVERSION:
      data[0] = 0x02;
      data[1] = 0x01;
      vdCommandType = CMDTYPE_SYSTEM;
      waitForAnswer = TRUE;
      answerType = ANSWERTYPE_SYSTEM;
    break;
    case VD_CMD_GETALLVERSION:
      data[0] = 0x02;
      data[1] = 0x01;
      vdCommandType = CMDTYPE_SYSTEM;
      waitForAnswer = TRUE;
      answerType = ANSWERTYPE_SYSTEM;
    break;
  }
  if(data[0] != 0)
  {
    vdSendData(data);
    hasToSendCommand = FALSE;
    delayMS(50);
    sendCmdRunSignal();
  }

  // Für Systembefehle soll eine Antwort sofort gewartet werden
  if(vdCommandType == CMDTYPE_SYSTEM)
  {
    canSendCommand = TRUE;
    if(waitForAnswer == TRUE)
    {
      waitForAnswer = FALSE;
      bzero(data, 10);
      waitForPINClearWithTimeOut(VD_PIN_SHS, VD_SHS, 5000); // 5 Sekunden TimeOut
      vdReceiveData(data);
      txbuffer[0] = ATMEGAANSWER_OTHER;
      if(cmd == VD_CMD_GETVERSION || cmd == VD_CMD_GETALLVERSION)
      { // Antwort ist festcodiert...
        for(p = 2, count = 0; count < 3; count++)
            txbuffer[p++] = data[count + 3];
        bzero(sTemp, 20);
        itoa(data[7], sTemp, 10); // Major Version
        for(l = 0; l < strlen(sTemp); l++)
          txbuffer[p++] = sTemp[l];
        txbuffer[p++] = '.';
        bzero(sTemp, 20);
        itoa(data[6], sTemp, 10); // Minor Version
        for(l = 0; l < strlen(sTemp); l++)
          txbuffer[p++] = sTemp[l];
        if(cmd == VD_CMD_GETALLVERSION)
        { // Programmsversionsnummer hinzufügen
          txbuffer[p++] = '-';
          bzero(sTemp, 20);
          strcpy(sTemp, VERSIONSTRING);
          for(l = 0; l < strlen(sTemp); l++)
            txbuffer[p++] = sTemp[l];
        }
        txbuffer[1] = p - 2; // Länge wird berechnet
      }
      else
      {
        for(p = 2, count = 3; count <= (data[1] + 1); count++)
          if(data[count] >= 'A' && data[count] <= 'Z')
            txbuffer[p++] = data[count];
          else
          {
            bzero(sTemp, 20);
            itoa(data[count], sTemp, 10);
            for(l = 0; l < strlen(sTemp); l++)
              txbuffer[p++] = sTemp[l];
          }
        txbuffer[1] = p - 3; // Länge wird berechnet
      }
      buffer_adr = 0;
      ATMEGA_PORT |= (1<<ATMEGA_SIGNAL) | (1<<ATMEGA_DATA);
      waitForACK();
    }
  }
}

/**
* Wird vom Hauptprogramm aufgerufen und erledigt alle Aufgaben, die in
* der Hauptschleife passieren sollen
*/
void doLoopJobsVoice()
{
  uint8_t data[10];
  int8_t responseCode, l, len;
  char sTemp[10];

  ATMEGA_PORT &= ~(1<<ATMEGA_SIGNAL);
  ATMEGA_PORT &= ~(1<<ATMEGA_DATA);

  if(hasToSendCommand == TRUE && canSendCommand == TRUE)
  {
    if(vdCommandToSend != VD_CMD_NOOP)
    {
      hasToSendCommand = FALSE;
      canSendCommand = FALSE;
      sendCommand(vdCommandToSend);
    }
    else
    {
      canSendCommand = TRUE;
      waitForAnswer = FALSE;
    }
  }

  if(waitForAnswer == TRUE && bit_is_set(VD_PIN_SHS, VD_SHS))
  {
    clearCmdRunSignal();
    waitForAnswer = FALSE;
    canSendCommand = TRUE;
    if(!vdReceiveData(data))
      return;

    responseCode = data[2];

    if(responseCode == RSP_RECOGNIZED || responseCode == RSP_MAYBERECOGNIZED)
    {
      data[0] = 0x02;
      data[1] = 0x26;
      vdSendData(data);
      if(!vdReceiveData(data))
        return;

      if(countCurMask != 0 && answerType == ANSWERTYPE_HEAR)
      {
        responseCode = RSP_NOTRECOGNIZED;
        for(l = 0; l < countCurMask; l++)
          if(data[3] == curMask[l])
          {
            responseCode = RSP_RECOGNIZED;
            break;
          }
      }
      if(responseCode == RSP_RECOGNIZED)
      {
        switch(answerType)
        {
          case ANSWERTYPE_HEAR:
            txbuffer[0] = ATMEGAANSWER_RECOGNIZED;
            txbuffer[1] = 1; // Länge = 1
            txbuffer[2] = data[3];
          break;
          case ANSWERTYPE_LEARN:
            txbuffer[0] = ATMEGAANSWER_LEARNED;
            txbuffer[1] = 1; // Länge = 1
            txbuffer[2] = data[3];
          break;
          case ANSWERTYPE_SYSTEM:
            txbuffer[0] = ATMEGAANSWER_OTHER;
            len = 1;
            if(data[3] >= 'A' && data[3] <= 'Z')
              txbuffer[2] = data[3];
            else
            {
              itoa(data[3], sTemp, 10);
              for(l = 0; l < strlen(sTemp); l++)
                txbuffer[(len++) + 1] = sTemp[l];
            }
            txbuffer[1] = len;
          break;
        }

        buffer_adr = 0;
        ATMEGA_PORT |= (1<<ATMEGA_SIGNAL) | (1<<ATMEGA_DATA);
        waitForACK();
      }
      else
      {
        switch(answerType)
        {
          case ANSWERTYPE_HEAR:
            txbuffer[0] = ATMEGAANSWER_NOTRECOGNIZED;
          break;
          case ANSWERTYPE_LEARN:
            txbuffer[0] = ATMEGAANSWER_LEARNED;
          break;
          case ANSWERTYPE_SYSTEM:
            txbuffer[0] = ATMEGAANSWER_OTHER;
          break;
        }
        txbuffer[1] = 0; // Länge = 0
        txbuffer[2] = 0; // Spaßdaten
        buffer_adr = 0;
        ATMEGA_PORT &= ~(1<<ATMEGA_DATA);
        ATMEGA_PORT |= (1<<ATMEGA_SIGNAL);
        waitForACK();
      }
    }
    else
    {
      switch(answerType)
      {
        case ANSWERTYPE_HEAR:
          txbuffer[0] = ATMEGAANSWER_NOTRECOGNIZED;
        break;
        case ANSWERTYPE_LEARN:
          txbuffer[0] = ATMEGAANSWER_LEARNED;
        break;
        case ANSWERTYPE_SYSTEM:
          txbuffer[0] = ATMEGAANSWER_OTHER;
        break;
      }
      txbuffer[1] = 0; // Länge = 0
      txbuffer[2] = 0; // Spaßdaten
      buffer_adr = 0;
      ATMEGA_PORT &= ~(1<<ATMEGA_DATA);
      ATMEGA_PORT |= (1<<ATMEGA_SIGNAL);
      waitForACK();
    }
  }
}
